/*
 *	Copyright (c) 1996 by Bill Putnam
 *      This Code is Freely Available, 
 *      Because Microsoft's Interface is More Complex than It Needs to Be
 *		All Rights Reserved.
 *
 *
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<math.h>
#include	<windows.h>
#include	<MMSystem.h>


#define TESTING  0

static void CALLBACK handleWaveOut(HWAVEOUT hwo, 
								  UINT uMsg, 
								  DWORD dwInstance, 
								  DWORD dwParam1, 
								  DWORD dwParam2);
void printErrorMessage( char* func, MMRESULT result );


extern int	paused;

 /* WinAudio Member Variables */
 int		playing;
 HWAVEOUT	audioPort;
 int		audioPortActive;
 int		bufferSize;
 int		outstandingBuffers;
 WAVEHDR	wh1, wh2;

 
 typedef short (*DataFunction)( short* );
 DataFunction myComputeFunc;


void InitializeSpeaker(	int channels, UINT sampleRate, UINT reqBufferSize, 
					   DataFunction computeFunc ){
	MMRESULT		result;
	WAVEFORMATEX	wfx;
	char			msg[256];
	
	myComputeFunc = computeFunc;

	bufferSize = reqBufferSize;

	audioPortActive = 0;
	playing = 0;
	audioPort = NULL;
	outstandingBuffers = 0;

	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nChannels = channels;
    wfx.nSamplesPerSec = sampleRate;
	wfx.nBlockAlign = sizeof(short) * channels;
    wfx.nAvgBytesPerSec =  sampleRate * wfx.nBlockAlign;
    wfx.wBitsPerSample = 8 * sizeof(short);
    wfx.cbSize = 0;

	result = waveOutOpen(&audioPort, 
						0, 
						&wfx, 
						(DWORD)handleWaveOut,
						(DWORD) NULL, 
						CALLBACK_FUNCTION);
	audioPortActive = 1;

	if (result != MMSYSERR_NOERROR)
		waveOutGetErrorText(result, msg, 256);

#if TESTING
	printErrorMessage("waveOutOpen", result);
#endif

	wh1.lpData = (LPSTR)calloc(bufferSize,sizeof(short));
	wh1.dwBufferLength = bufferSize * sizeof(short);
	wh1.dwBytesRecorded = 0;
	wh1.dwUser = 1;
	wh1.dwFlags = 0;
	wh1.dwLoops = 0;
	wh1.lpNext = NULL;
	wh1.reserved = 0;

	wh2.lpData = (LPSTR)calloc(bufferSize,sizeof(short));
	wh2.dwBufferLength = bufferSize * sizeof(short);
	wh2.dwBytesRecorded = 0;
	wh2.dwUser = 2;
	wh2.dwFlags = 0;
	wh2.dwLoops = 0;
	wh2.lpNext = NULL;
	wh2.reserved = 0;


}

/* Prepare 2 headers, and send out the 1st 2 buffers of data 
 */
void StartPlaying(){
	MMRESULT	result;

	playing = 1;

	/* set up the first 2 buffers */
	result = waveOutPrepareHeader(audioPort, &wh1, sizeof(WAVEHDR));

	#ifdef TESTING
		printErrorMessage("waveOutPrepareHeader1", result);
    #endif

	result = waveOutPrepareHeader(audioPort, &wh2, sizeof(WAVEHDR));
	
	#ifdef TESTING
		printErrorMessage("waveOutPrepareHeader2", result);
    #endif
	
 	outstandingBuffers++;
	result = waveOutWrite(audioPort, &wh1, (UINT)sizeof(WAVEHDR));
	

	#ifdef TESTING
		printErrorMessage("waveOutWrite1:", result);
    #endif
	
	outstandingBuffers++;
	result = waveOutWrite(audioPort, &wh2, (UINT)sizeof(WAVEHDR));
	

	#ifdef TESTING
		printErrorMessage("waveOutWrite2:", result);
    #endif

	

}

void StopPlaying(void) 
{

	if(playing) 
		playing = 0;

	audioPortActive = 0;

}

void CloseAudioOutput(){

	MMRESULT result;

	
	if (audioPort != NULL) {

		result = waveOutUnprepareHeader(audioPort, 
										   &wh1,
										   sizeof(WAVEHDR));	
		result = waveOutUnprepareHeader(audioPort, 
										   &wh2,
										   sizeof(WAVEHDR));	
		
		if (wh1.lpData != NULL) {
			free(wh1.lpData);
			wh1.lpData = NULL;
		}
		if (wh2.lpData != NULL) {
			free(wh2.lpData);
			wh2.lpData = NULL;
		}		

		result = waveOutClose(audioPort);
	
	}
}

/*
int GetSampleRate() {
	return kDefaultSampleRate;
}
	

*/	


/*******************************************************/
/** Test Routines */

void printErrorMessage( char* func, MMRESULT result ){
	char msg[256];
	waveOutGetErrorText(result, msg, 256);
	printf("%s: %s\n",func, msg);
}


/////////////////////////////////////////////////////////////////////////////
// Audio Handler
//
   	
static void CALLBACK handleWaveOut(HWAVEIN hwi, 
								  UINT uMsg, 
								  DWORD dwInstance, 
								  DWORD dwParam1, 
								  DWORD dwParam2)
{
		
	switch (uMsg) {
	case WOM_OPEN:
		#ifdef TESTING_VERBOSE
			printf("WOM_OPEN\n");
		#endif
		break;
	case WOM_CLOSE:
		#ifdef TESTING_VERBOSE
			printf("WOM_CLOSE\n");
		#endif
		break;
	case WOM_DONE: 
		{
			MMRESULT result;
  			WAVEHDR *pwh = (WAVEHDR *)dwParam1;
			
			
			result = waveOutUnprepareHeader(audioPort, pwh, sizeof(WAVEHDR));
			

			if (playing) {
					
				if (!((*myComputeFunc)( (short*)pwh->lpData ) ) )
					playing = 0;

				result = waveOutPrepareHeader(audioPort, pwh, sizeof(WAVEHDR));
				result = waveOutWrite(audioPort, pwh, sizeof(WAVEHDR));
				result++;
			} else {
				outstandingBuffers--;
				if (outstandingBuffers == 0)
					CloseAudioOutput();
					
			}
		}
 		break;
	}
}


/*****************************************************
 * 
 *			* WinSpkr Test Code *
 *
 *****************************************************/


#if TESTING

#include <conio.h>
#define mySampleRate 22050
#define myBufferSize 10*1024
#define kChannels 1

short			*testData1,*testData2;

short dataFunction( short* myData) {
	static BOOL state = FALSE;

	if (!state)
		memcpy(myData, testData1, myBufferSize*sizeof(short)*kChannels );
	else
		memcpy(myData, testData2, myBufferSize*sizeof(short)*kChannels );

	state = !state;

	return 1;
}

void main( int argc, char **argv ){

	int i;
	
	testData1 = (short*)malloc(myBufferSize*sizeof(short));
	testData2 = (short*)malloc(myBufferSize*sizeof(short));
	for (i=0; i<myBufferSize; i++) {
		testData1[i] = (short)(30000*sin(6.28*(200.6463) /mySampleRate*i));
		testData2[i] = (short)(30000*sin(6.28*(200.6463)/mySampleRate*(i+myBufferSize)));
	}

	InitializeSpeaker( 1, (UINT)22050, (UINT)myBufferSize, dataFunction );
	StartPlaying();
	
	
	while( !kbhit() )
		;
	getch();

	StopPlaying();


	if(testData1 != NULL)
			free(testData1);
	if(testData2 != NULL)
			free(testData2);

}

#endif /* TESTING */
